import enum
from dataclasses import dataclass
from typing import Protocol

from sqlalchemy.pool import StaticPool


class DriverTypes(Protocol):
    port: int
    has_config: bool
    async_driver: str
    sync_driver: str

    config = {}

    def get_connection_uri(self, is_async: bool, cfg: 'DatabaseConfig') -> str:
        """Returns autogenerated driver uri for sqlalchemy create_engine()"""
        driver_prefix = self.async_driver if is_async else self.sync_driver
        return f'{driver_prefix}://{cfg.user}:{cfg.passwd}@{cfg.host}:{cfg.get_port()}/{cfg.name}'


class MysqlDriver(DriverTypes):
    """Driver Default values for MySQL Connection"""

    port = 3306
    async_driver = 'mysql+aiomysql'
    sync_driver = 'mysql+pymysql'
    has_config = False


class PostgresDriver(DriverTypes):
    """Driver Default values for PostgreSQL Connection"""

    port = 5432
    async_driver = 'postgresql+asyncpg'
    sync_driver = 'postgresql+psycopg2'
    has_config = False


class SqliteDriver(DriverTypes):
    port = 0
    async_driver = 'sqlite+aiosqlite'
    sync_driver = 'sqlite'
    has_config = True

    config = {
        'connect_args': {'check_same_thread': False},
        'poolclass': StaticPool,
    }

    def get_connection_uri(self, is_async: bool, cfg: 'DatabaseConfig') -> str:
        driver_prefix = self.async_driver if is_async else self.sync_driver
        return '{driver}:///{host}'.format(driver=driver_prefix, host=cfg.host)


class Driver(str, enum.Enum):
    MYSQL = 'mysql'
    POSTGRES = 'postgres'
    SQLITE = 'sqlite'


@dataclass(frozen=True)
class DatabaseConfig:
    """Database configuration params
    Obs: pass filename as name if using sqlite"""

    driver: DriverTypes
    host: str
    name: str = ''
    user: str = ''
    passwd: str = ''
    port: int = -1
    _pool_size: int = 20
    _pool_recycle: int = 3600
    _max_overflow: int = 0

    def get_port(self):
        return self.port if self.port != -1 else self.driver.port

    def get_uri(self, *, is_async: bool):
        return self.driver.get_connection_uri(is_async, self)

    @property
    def pool_config(self) -> dict[str, int]:
        if self.driver.has_config:
            return self.driver.config
        return {
            'pool_size': self._pool_size,
            'pool_recycle': self._pool_recycle,
            'max_overflow': self._max_overflow,
        }

    @classmethod
    def new(cls, driver_type: Driver, **kwargs):
        _driver_mapping = {
            Driver.MYSQL: MysqlDriver,
            Driver.POSTGRES: PostgresDriver,
            Driver.SQLITE: SqliteDriver,
        }
        return cls(driver=_driver_mapping[driver_type](), **kwargs)



from contextlib import asynccontextmanager
from typing import AsyncGenerator

from context_handler import AsyncContext
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker

from souswift_core.providers.database.base import DatabaseProvider


class SessionProvider(DatabaseProvider):
    def get_session(self):
        return sessionmaker(
            bind=self.engine, expire_on_commit=False, class_=AsyncSession
        )()

    @asynccontextmanager
    async def acquire(self) -> AsyncGenerator[AsyncSession, None]:
        async with self.get_session() as session:   # type: AsyncSession
            async with session.begin():
                yield session

    @staticmethod
    def is_closed(client: AsyncSession):
        return not client.is_active

    @staticmethod
    async def close_client(client: AsyncSession):
        await client.close()

    async def health_check(self):
        async with self.acquire() as session:
            await session.execute(text('SELECT 1'))
        self.logger.info('Database Connection succeeded')

    Context = AsyncContext[AsyncSession]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['souswift_core',
 'souswift_core.filters',
 'souswift_core.providers',
 'souswift_core.providers.database']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.29,<2.0.0',
 'aiomysql>=0.0.22,<0.0.23',
 'context-handler>=2.1.0,<3.0.0',
 'starlette>=0.16.0,<0.17.0']

setup_kwargs = {
    'name': 'souswift-core',
    'version': '0.4.2',
    'description': '',
    'long_description': None,
    'author': 'Gustavo Correa',
    'author_email': 'self.gustavocorrea@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
